/*
 *  dosgate 1.0, a Linux+dosemu interface to dos-based screen readers
 *  Copyright (C) 1999  Enrico Zini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  You can reach the author at zinie@cs.unibo.it
 */
#include <stdio.h>
#include <stdlib.h>
#include <bios.h>
#include <dos.h>

#include <mem.h>

#include "bool.h"
#include "interrupt.h"
#include "keyboard.h"
#include "cmdq.h"

/* Maximum size in words of BIOS keyboard queue */
#define MAX_BIOS_KQ_SIZE 256

#define SHADOW

/* 'inline' functions */
#define enqueue_vc_change(vc_no) cmd_enqueue(CMD_VC_CHANGE, (vc_no))
#define enqueue_scancode(scancode) cmd_enqueue(CMD_SCANCODE, (scancode))
#define raw_peek_char() inportb(0x60)
#define end_of_hw_interrupt() outportb(0x20, 0x20)
#define raw_consume_char() do {	\
				unsigned char c = inportb(0x61); \
				outportb(0x61, c | 0x80); \
				outportb(0x61, c & 0x7f); \
			   } while(0)

extern int_handler old_int09_handler;
extern int_handler old_int16_handler;

struct kb_flags {
	unsigned int extended:1;
	unsigned int lt_alt:1;
	unsigned int rt_alt:1;
};

struct kq {
	unsigned short int head;
	unsigned short int tail;
	unsigned short int size;
	unsigned short int data[MAX_BIOS_KQ_SIZE];
};

static unsigned short int *bkq_start = MK_FP(0x40, 0x80);
static unsigned short int *bkq_end = MK_FP(0x40, 0x82);
static unsigned short int *bkq_next = MK_FP(0x40, 0x1a);
static unsigned short int *bkq_free = MK_FP(0x40, 0x1c);

static struct kq shadow_kq;

/*
static void make_bios_kq_snapshot()
{
	shadow_kq.head = kq_state->head;
	shadow_kq.tail = kq_state->tail;
	shadow_kq.size = kq_pos->end - kq_pos->start;
	movedata(0x40, kq_pos->start,
		 FP_SEG(shadow_kq.data), FP_OFF(shadow_kq.data),
		 shadow_kq.size);
}
*/

/* FIXME: that kb_flags structure should perhaps be module global */
static bool handle_function_press(struct kb_flags f, unsigned char n)
{
	if (f.lt_alt)
		enqueue_vc_change(n);
	else if (f.rt_alt)
		enqueue_vc_change(12+n);
	else
		return FALSE;

	return TRUE;
}

#define K_CODE(c) ((c) & 0x7f)
#define K_PRESS(c) (1-(((c) >> 7) & 1))
/* Intercettare invece INT 15 AH=0x4f */
void interrupt int09_handler()
{
	static struct kb_flags flags = {0, 0, 0};
	bool processed = FALSE;
	unsigned char c = raw_peek_char();
	unsigned short int skn = *bkq_next;
	unsigned short int skf = *bkq_free;

	if (flags.extended)
	{
		flags.extended = 0;

		if (K_CODE(c) == 0x38) /* Right alt */
			flags.rt_alt = K_PRESS(c);
	}
	else
	{
		if (c == 0xe0)
			flags.extended = 1;
		else if (K_CODE(c) == 0x38) /* Left alt */
			flags.lt_alt = K_PRESS(c);
		else if (K_CODE(c) == 0x10) /* q */
		{
			if (flags.rt_alt && flags.lt_alt)
			{
				cmd_enqueue(CMD_QUIT, 0);
				processed = TRUE;
				enqueue_scancode(0xb8);
				enqueue_scancode(0xe0);
				enqueue_scancode(0xb8);
			}
		}
		else if (K_PRESS(c) && K_CODE(c) >= 0x3b && K_CODE(c) <= 0x44)
			/* F1..F10 */
			processed = handle_function_press(flags, K_CODE(c)-0x3a);
		else if (K_PRESS(c) && K_CODE(c) >= 0x57 && K_CODE(c) <= 0x58)
			/* F11..F12 */
			processed = handle_function_press(flags, 10+K_CODE(c)-0x56);
	}

	if (processed)
	{
		raw_consume_char();
		end_of_hw_interrupt();
	}
	else
	{
		enqueue_scancode(c);
		old_int09_handler();
		*bkq_next = skn;
		*bkq_free = skf;
		enable();
		return;
	}
}

bool shadow_haskey()
{
	return shadow_kq.head != shadow_kq.tail;
}

int shadow_getkey()
{
	int res;
	
	while(!shadow_haskey())
		;

	if (shadow_kq.head < 0 || shadow_kq.head > MAX_BIOS_KQ_SIZE)
		return 0;

	res = shadow_kq.data[shadow_kq.head]; 
	shadow_kq.head = (shadow_kq.head+1) % shadow_kq.size;

	return 0xf000 | shadow_kq.head;
}

/*
static unsigned short int *bkq_start = MK_FP(0x40, 0x80);
static unsigned short int *bkq_end = MK_FP(0x40, 0x80);
static unsigned short int *bkq_next = MK_FP(0x40, 0x1a);
static unsigned short int *bkq_free = MK_FP(0x40, 0x1c);
 */

void my_bios_stats()
{
	printf("next: %x, free: %x, start: %x, end: %x\n", *bkq_next, *bkq_free, *bkq_start, *bkq_end);
}

bool my_bios_haskey()
{
	return *bkq_next != *bkq_free;
}

int my_bios_getkey()
{
	int res;
	
	while (!my_bios_haskey())
		;

	disable();
	res = *(int far *)MK_FP(0x40, *bkq_next);/*kq_pos->start+kq_state->head);*/
	(*bkq_next) += 2;
	if (*bkq_next == *bkq_end)
		*bkq_next = *bkq_start;
	enable();
	/*kq_state->head = (kq_state->head+1) % (kq_pos->end - kq_pos->start);*/
		
	return res;
}

int bios_peek_extended()
{
	union REGS in, out;
	in.h.ah = 0x11; /* FIXME: use the right number */
	int86(0x16, &in, &out);
	return out.x.ax;
}

int bios_get_extended()
{
	union REGS in, out;
	in.h.ah = 0x10; /* FIXME: use the right number */
	int86(0x16, &in, &out);
	return out.x.ax;
}

bool bios_haskey()
{
#ifdef NONE
	union REGS in, out;
	in.h.ah = 0x10; /* FIXME: use the right number */
	int86(0x16, &in, &out);
	return out.x.ax;
#endif
#ifdef SHADOW
	/*return shadow_haskey();*/
	return my_bios_haskey();
	/*return bioskey(1) != 0;*/
#else
	return bioskey(1) != 0;
#endif
}

int bios_getkey()
{
#ifdef SHADOW
	/* Mangia un tasto dal BIOS per non incepparlo */
	/*
	if (bioskey(1))
		bioskey(0);
	return shadow_getkey();
	*/
	return my_bios_getkey();
	/*return bioskey(0);*/
#else
	union REGS in, out;
	in.h.ah = 0x10;
	int86(0x16, &in, &out);
	return out.x.ax;
#endif
}

void interrupt int16_handler()
{
/*
	if (_AH == 0x05)
		cmd_enqueue(CMD_BIOS_INS, _CH);
	else
		old_int16_handler();
*/
/*
	if (_AH == 0x05)
		cmd_enqueue(CMD_BIOS_INS, _CH);
	else
		old_int16_handler();

	if (_AH != 0x01)
		cmd_enqueue(CMD_BIOS_INS, _AH);
*/
}
