/*
 *  dosgate 1.0, a Linux+dosemu interface to dos-based screen readers
 *  Copyright (C) 1999  Enrico Zini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  You can reach the author at zinie@cs.unibo.it
 */
#include <stdio.h>
#include <io.h>
#include <sys\stat.h>
#include "lingate.h"

/* Struct of a vkeyboard command */
#pragma pack(1)
struct vk_cmd
{
	unsigned long int vc_num;
	unsigned long int scancode;
};
#pragma pack()

static long int current_vc = -1;	/* Current virtual console number */
static long int old_vc = -1;		/* Previous virtual console number */
static int vcsa_fd = -1;	/* Current virtual console's fd */
static int vkeyboard_fd = -1;	/* vkeyboard device fd */
static const char *dev_path;	/* Directory containing Linux vcsa and vkeyboard devices */
static char fname[255];		/* Buffer space to compose file names */

/* Initialize lingate module */
int lg_init(const char *ldev_path)
{
	int res = 0;
	
	dev_path = ldev_path;

	sprintf(fname, "%s\\vcsa1", dev_path);
	if (access(fname, 0) == 0)
		res |= LG_HAS_VCSA;
	else
	{
		fprintf(stderr, "lg_init: failed vcsa support check on file %s", fname);
		perror("");
	}

	sprintf(fname, "%s\\vkbd", dev_path);
	vkeyboard_fd = open(fname, O_WRONLY);
	if (vkeyboard_fd != -1)
		res |= LG_HAS_VKBD;
	else
	{
		fprintf(stderr, "lg_init: failed vkeyboard support check on file %s", fname);
		perror("");
	}

	return res;
}

/* Shut down lingate module */
void lg_end()
{
	if (vcsa_fd != -1)
		close(vcsa_fd);

	if (vkeyboard_fd != -1)
		close(vkeyboard_fd);
}

long int lg_current_vc()
{
	return current_vc;
}


/* Change the current vc to vc_no */
void lg_switch_vc(long int vc_no)
{
	/* We use globals, so we are not reentrant anyway */
	static char fname[255];
	int new_fd;

	sprintf(fname, "%s\\vcsa%d", dev_path, vc_no);
	new_fd = open(fname, O_RDONLY | O_BINARY);
	if (new_fd != -1)
	{
		if (vcsa_fd != -1)
			close(vcsa_fd);
		vcsa_fd = new_fd;
		old_vc = current_vc;
		current_vc = vc_no;
	}
}

/* Read Linux's vcsa screen; we need to be fast, so avoid checking for errors */
void lg_read_screen(struct coord *geometry, struct coord *cursor, unsigned char *content)
{
	lseek(vcsa_fd, 0, SEEK_SET);
	read(vcsa_fd, geometry, sizeof(struct coord));
	read(vcsa_fd, cursor, 2);
	read(vcsa_fd, content, geometry->x * geometry->y * 2);
}

/* Forward scancode to vkeyboard device */
void lg_forward_scancode(unsigned char scancode)
{
	struct vk_cmd c;
	c.vc_num = current_vc > 0 ? current_vc : 2;
	c.scancode = (unsigned long int)scancode;
	write(vkeyboard_fd, &c, sizeof(struct vk_cmd));
}
