/*
 *  vkeyboard 1.0, a device to feed raw keyboard scancodes to a virtual console
 *  Copyright (C) 1999  Enrico Zini and Nicolas Pitre
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  You can reach the authors at zinie@cs.unibo.it or nico@cam.org
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kbd_ll.h>
#include <asm/uaccess.h>


#ifndef PDEBUG 
#  ifdef DEBUG_modulename 
#    ifdef __KERNEL__ 
#      define PDEBUG(fmt, args...) printk (KERN_DEBUG fmt "\n", ## args) 
#    else 
#      define PDEBUG(fmt, args...) fprintf (stderr, fmt "\n", ## args) 
#    endif 
#  else 
#    define PDEBUG(fmt, args...) 
#  endif 
#endif 
 
#ifndef PDEBUGG 
#  define PDEBUGG(fmt, args...) 
#endif


#ifndef DEVICE_NAME
#define DEVICE_NAME "vkeyboard"
#endif


struct vk_cmd
{
	unsigned long vc_num;
	unsigned long scancode;
};


static int vkeyboard_major;

#ifdef MODULE
#ifdef FG_CONSOLE_ADDR
#define FG_CONSOLE *(int *)FG_CONSOLE_ADDR
#else
#warning FG_CONSOLE_ADDR is not defined compiling as a module, insmod will probably fail.
extern int fg_console
#define FG_CONSOLE fg_console
#endif
#else
extern int fg_console;
#define FG_CONSOLE fg_console
#endif


static ssize_t vkeyboard_write( struct file *file, const char *buf,
				size_t count, loff_t *ofs)
{

	PDEBUG("vkeyboard write");

	if (count % sizeof(struct vk_cmd) != 0)
	{
		printk("Wrong write on vkeyboard device (count=%i)!\n", count);
		return 0;
	}
	else
	{
		struct vk_cmd cmd;
		int i;
		for (i = 0; i < count; i += sizeof(struct vk_cmd))
		{
			int saved_console = FG_CONSOLE;

			copy_from_user(&cmd, buf+i, sizeof(cmd));

			if (cmd.vc_num > 0 && cmd.vc_num < 24)
			{
				printk("Scancode %lx wrote to vc %lx\n", 
					cmd.scancode, cmd.vc_num-1 );

				FG_CONSOLE = cmd.vc_num - 1;
				handle_scancode(cmd.scancode/* & 0x7f*/, !(cmd.scancode & 0x80)/*>> 7*/);
				FG_CONSOLE = saved_console;
			} else {
				printk("Scancode %lx sent to invalid console %lx\n",
						cmd.scancode, cmd.vc_num-1);
				return 0;
			}
		}
	}

	return count;
}


static int vkeyboard_open(struct inode *inode, struct file *file) 
{
	PDEBUG("vkeyboard open");
	MOD_INC_USE_COUNT;
	return 0;
}


static int vkeyboard_close(struct inode *inode, struct file *file) 
{
	PDEBUG("vkeyboard release");
	MOD_DEC_USE_COUNT;
	return 0;
}


static struct file_operations vkeyboard_fops = {
	NULL,			/* llseek */
	NULL,			/* read */
	vkeyboard_write,	/* write */
	NULL,			/* readdir */
	NULL,			/* poll */
	NULL,			/* ioctl */
	NULL,			/* mmap */
	vkeyboard_open,		/* open */
	NULL,			/* flush */
	vkeyboard_close,	/* release */
	NULL,			/* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL,			/* revalidate */
	NULL 			/* lock */
};


int vkeyboard_init( void ) 
{
	PDEBUG("vkeyboard init");
	vkeyboard_major = register_chrdev(0, DEVICE_NAME, &vkeyboard_fops);
	if( vkeyboard_major < 0 ){
		printk("unable to get major for vkeyboard device\n");
		return -EIO;
	}
	return 0;
}


#ifdef MODULE
int init_module()
{
	return vkeyboard_init();
}

void cleanup_module()
{
	unregister_chrdev(vkeyboard_major, DEVICE_NAME);
}
#endif
